from file import read_JSON, read_file, write_csv, prettyPrint
import random, math, time, os, json
import multiprocessing as mp
from copy import deepcopy
import numpy as np, os
from sklearn import mixture, preprocessing
from sklearn.externals import joblib
from statsmodels.sandbox.stats.multicomp import multipletests
from operator import itemgetter
from statistics import generateBackgroundModel, generateScore, generatePVals, compoundSetGSEA, scoreRankedLists, makeMixmdl
from CGEAdata import CGEAdata
from featureListProcessing import formatGeneLists

def featureBasedCGEA(upgenes, downgenes, saveAs, outputDir, sortDir, data, status, alpha = .05, numWorkers = mp.cpu_count(), pool = None, verbose = True):
    #print 'Start - 0'
    status[1] = 'Started'
    status[2] = 0.01
    curTime = time.time()
    data = CGEAdata(data = data)
    status[1] = 'Loading cmap data'
    if verbose:
        print status[1]
    cmap = data.get('cmap')
    #print "Loaded cmap data - took " + str(time.time()-curTime) + ' seconds'
    status[1] = 'Loaded cmap data'
    if verbose:
        print status[1]
    status[2] = 0.03
    
    geneLists = formatGeneLists([upgenes, downgenes])
    
    status[1] = 'Generating Null Model'
    curTime = time.time()
    [backgroundModel, backgroundScores, scaler] = generateBackgroundModel(geneLists, cmap, status=status, numWorkers = numWorkers)
    
    #print "Generated Null Model - took " + str(time.time()-curTime) + ' seconds'
    status[1] = 'Generated Null Model'
    if verbose:
        print status[1]
    status[2] = 0.33
    curTime = time.time()
    
    status[1] = 'Scoring Compounds'
    if verbose:
        print status[1]
#     drugIndex = {}
#     scores = []
#     index = {}
#     countCMAP = len(cmap.keys())
#     for count, value in enumerate(cmap.items()):
#         if status:
#             status[2] += 1.0/(countCMAP)*3.3/10.0
#         name = value[0]
#         cmapList = value[1]
#         #if (count+1)%100 == 0: print str(count+1)+" cmap drugs scored"
#         score = generateScore(geneLists, cmapList)[0]
#         index[count] = name
#         drugIndex[name] = count
#         scores.append([score])
    
    [drugIndex, scores, index] = scoreRankedLists(geneLists, cmap, status=status, numWorkers = numWorkers, pool = pool)
    
    #TESTING
    #drugList = []
    #for drug in drugIndex.keys():
    #    index = drugIndex[drug]
    #    drugList.append([drug,scores[index][0]])
    #print sorted(drugList, key = itemgetter(1), reverse = False)[0:10]
    #TESTING
    
    #print sorted([[x, scores[drugIndex[x]]] for x in drugIndex.keys()], key = itemgetter(1))[0:10]
    #print "Generated Annotated Compound Scores - took " + str(time.time()-curTime) + ' seconds'
    curTime = time.time()
    [pvals, adjpvals, scores] = generatePVals(backgroundModel, scores, backgroundScores = backgroundScores, scaler = scaler, alpha = alpha, numWorkers = numWorkers, pool = pool)
    #print "Generated Annotated Compound Pvals - took " + str(time.time()-curTime) + ' seconds'
    
    
    #TESTING
    #drugList = []
    #for drug in drugIndex.keys():
    #    index = drugIndex[drug]
    #    drugList.append([drug,scores[index]])
    #print sorted(drugList, key = itemgetter(1), reverse = False)[0:10]
    #raise hell
    #TESTING
    
    
    status[1] = 'Compounds Scored and Annotated'
    if verbose:
        print status[1]
    status[2] = 0.66
    curTime = time.time()
    drugAnnotations = {}
    for drug in drugIndex.keys():
        index = drugIndex[drug]
        drugAnnotations[drug]={'Name':drug,
                               'Score':scores[index],
                               'P-Value':pvals[index],
                               'Adjusted P-Value':adjpvals[index]
                               }
    #print '\n'.join(map(lambda x: ' '.join([x['Name'], str(x['Adjusted P-Value']), str(x['Score'])]),sorted(drugAnnotations.values(), key = itemgetter('Score'))[0:10]))
    status[1] = 'Performing Gene Enrichment'
    if verbose:
        print status[1]
    [rankedAnnotatedDrugs, compoundSetResults] = geneEnrichment(drugAnnotations, data, status = status, numWorkers = numWorkers, pool = pool)
    results = {'Annotated Compounds':rankedAnnotatedDrugs, 'Combined Enrichment Results':compoundSetResults}
    #print "Performed Gene Enrichment - took " + str(time.time()-curTime) + ' seconds'
    status[1] = 'Performed Gene Enrichment'
    if verbose:
        print status[1]
    status[2] = 0.99
    curTime = time.time()
    return results
    
      
typos={'X6.bromoindirubin.3..oxime':'6-bromoindirubin-3-oxime',
       'alpha.estradiol':'alpha-estradiol'}
# sorting = {'up':lambda x: sorted(list(x.values()), key = itemgetter('Score')),
#            'down': lambda x: sorted(list(x.values()), key = itemgetter('Score')),
#            'abs': lambda x: sorted(list(x.values()), key = lambda x: abs(x['Score']))
#            }
mySort = {'up':lambda x: x.sort(key = itemgetter('Score')),
          'down': lambda x: sort(key = itemgetter('Score'), reverse = True),
          'abs': lambda x: x.sort(key = lambda y: abs(y['Score']))
          }
def geneEnrichment(drugAnnotations, data, sortDir = 'up', status = None, numWorkers = mp.cpu_count(), pool = None):
    
    #######################
    # Annotated Compounds #
    #######################
    curTime = time.time()
    cmapDrugMetaData = data.get("cmapDrugMetaData")
    for drugData in cmapDrugMetaData:
        if drugData['name'] in drugAnnotations: 
            #print 'Rname'
            drugAnnotations[drugData['name']].update(drugData)
            del drugAnnotations[drugData['name']]['name']
        elif drugData['Rname'] in drugAnnotations: 
            #print 'name'
            drugAnnotations[drugData['Rname']].update(drugData)
            del drugAnnotations[drugData['Rname']]['name']
        elif drugData['name'] in typos:
            drugAnnotations[typos[drugData['name']]].update(drugData)
            del drugAnnotations[typos[drugData['name']]]['name']
        elif drugData['Rname'] in typos:
            drugAnnotations[typos[drugData['Rname']]].update(drugData)
            del drugAnnotations[typos[drugData['Rname']]]['name']
        else:
            print 'Missing from annotations: '+ drugData['name'] + '/' + drugData['Rname']
    del cmapDrugMetaData
    #print time.time() - curTime
    #curTime = time.time()
    
    rankedAnnotatedDrugs = drugAnnotations.values()
    mySort[sortDir](rankedAnnotatedDrugs)
    
    sizeThreshold = 5
    targetEnrSeedPAdj = .1
    minCompoundSetSizeThreshold = 5
    
    #######################
    # Combined Enrichment #
    #######################
    unifiedCompoundLevelSets = data.get("unifiedCompoundLevelSets")
    mixmdls = data.getModel('mixmdl')
    
    if not mixmdls:
        #If this has been reached, you are trying to run this without loading the data from bitbucket.
        print 'Gene Enrichment Mixture Models are not loaded. Computing Compound Set Mixmdls.'
        setSizes = reduce(lambda x,y: set.union(x,set(map(len,y.values()))), unifiedCompoundLevelSets.values(), set())
        #print maxSetSize
        makeMixmdl(data.get('cmap').keys(), data.getPath('mixmdl'), setSizes = setSizes, numIterations = 1000)
        mixmdls = data.getModel('mixmdl')
    #print time.time() - curTime
    #curTime = time.time()
    status[2]+=.03
    enrichmentData = data.getEnrichmentData()
    rankedDrugs = map(itemgetter('Name'),rankedAnnotatedDrugs)
    compoundSetResults = {}
    numSets = len(unifiedCompoundLevelSets.keys()+enrichmentData.keys())
    for setType, compoundSets in unifiedCompoundLevelSets.items()+enrichmentData.items():
        compoundSets = filter(lambda x: len(x[1])>minCompoundSetSizeThreshold, compoundSets.items())
        if len(compoundSets) > 0:
            compoundSetResults[setType] = compoundSetGSEA(rankedDrugs, mixmdls, compoundSets, numWorkers = numWorkers, pool = pool)
            #print 'Generated Results for ' + setType
        status[2] += 1.0/numSets*3.0/10.0
    
    #print time.time() - curTime
    #Generate Summary
    summaryResults = reduce(lambda x,y: x+map(lambda z: dict(zip(['Compound Set'] + z.keys(), [y[0]]+z.values()))
                                                        ,y[1])
                                      ,compoundSetResults.items(), [])
    mySort[sortDir](summaryResults)
    return [rankedAnnotatedDrugs, summaryResults]



from copy import deepcopy
from file import read_JSON
import os

class drugListProcessing():
    typesOfID = ["CAS","chembl_id","chebi_id","DrugBank","PubChem","MeSH_ID","KEGG_drug","name","Rname","stitch_id"]
    namePrettifier = {'name':'Name'}#,
#                       'target_entrez':'Target Entrez IDs',
#                       'union_target':'Union Targets',
#                       'level2_description': 'Level 2 Description',
#                       'level1_description': 'Level 1 Description',
#                       'chembl_id':'Chembl ID',
#                       'SEAtarget_entrez': 'SEA Target Entrez IDs',
#                       'chebi_id': 'Chebi ID',
#                       'level4_description': 'Level 4 Description',
#                       'moa':'Method of Action',
#                       'union_entrez': 'Union Entrez IDs',
#                       'target': 'Targets',
#                       'KEGG_drug': 'KEGG Drug ID',
#                       'atc_code':'ATC Code',
#                       'indication':'Indication',
#                       'level3_description': 'Level 3 Description',
#                       'side_effects': "Side Effects",
#                       'SEAtarget': "SEA Target"}
    bannedCols = ['Rname']
    def __init__(self, cmap):
        if isinstance(cmap, basestring) and os.path.isfile(cmap):
            cmap = read_JSON(cmap)
        self.cmap = cmap
        allDrugIDsByType = dict(map(lambda x: (x,
                                         set(map(lambda y: y[x].lower(),
                                                 filter(lambda y: x in y,
                                                        cmap)
                                                 ))
                                         ),
                              self.typesOfID)
                          )
        entryLookupByID = {}
        for entry in cmap:
            for type in self.typesOfID:
                if type in entry and entry[type] not in ['NA', '']:
                    if not entry[type].lower() in entryLookupByID:
                        entryLookupByID[entry[type].lower()] = entry
                    elif entryLookupByID[entry[type].lower()] == entry:
                        pass #normal
                    else:
                        countNA_1 = entry.values().count('NA')
                        countNA_2 = entryLookupByID[entry[type].lower()].values().count('NA')
                        if countNA_1 < countNA_2:
                            entryLookupByID[entry[type].lower()] = entry
                        elif countNA_1 == countNA_2:
                            #print entry[type] + ' was duplicated in cmap for a different entry and no entry was determined to be better.'
                            pass
            for key, value in entry.items():
                if key in self.namePrettifier:
                    del entry[key]
                    entry[self.namePrettifier[key]] = value
            del entry['Rname']
                            
        self.entryLookupByID = entryLookupByID
        #print allDrugIDsByType.keys()
        allDrugIDs = set.union(*allDrugIDsByType.values())
        #print len(allDrugIDs)
        self.allDrugIDsByType = allDrugIDsByType
        self.allDrugIDs = allDrugIDs
    
    def inCMAP(self, ids, type = None):
        if type and not type in self.allDrugIDsByType:
            print str(type) + ' is not in the list of accepted types.'
            return False
        ids = deepcopy(ids)
        if (isinstance(ids, list) or isinstance(ids, set)) and isinstance(next(iter(ids)), basestring):
            for index, id in enumerate(ids):
                if not type:
                    ids[index] = (id, id.lower() in self.allDrugIDs)
                else:
                    ids[index] = (id, id.lower() in self.allDrugIDsByType[type])
            return ids
        elif isinstance(ids, basestring):
            if not type:
                return ids.lower() in self.allDrugIDs
            else:
                return ids.lower() in self.allDrugIDsByType[type]
        else:
            if (isinstance(ids, list) or isinstance(ids, set)):
                print str(type(next(iter(ids)))) + ' is not a valid input type for id'
            else:
                print str(type(ids)) + ' is not a valid input type for id'
        return False
        
    def annotateIDs(self, ids):
        drugList = self.inCMAP(ids)
        hits = map(lambda x: x[0], filter(lambda x: x[1],drugList))
        misses = map(lambda x: x[0], filter(lambda x: not x[1],drugList))
        for index, hit in enumerate(hits):
            hits[index] = self.namePrettifier[self.entryLookupByID[hit.lower()]]
        return (hits, misses)
    
    def annotateCMAP(self, hits):
        if not isinstance(hits, set):
            hits = set(hits)
        for row in self.cmap:
            if row['Name'] in hits:
                row['In Query Set'] = 'True'
            else:
                row['In Query Set'] = 'False'
            
        return self.cmap


# set of utilities to interact with files

# @author: rm3086 (at) columbia (dot) edu

import csv, shutil, os, glob, cPickle, jsonpickle, json
from log import slogger

log = slogger ('CGEA')

# read an object (list, dictionary, set) from a serialized file
def read_obj (filename):
	try:
		data = cPickle.load (open(filename, 'rb'))
		return data
	except Exception as e:
		log.error(e)
		return None


# write an object (list, set, dictionary) to a serialized file
def write_obj (filename, data):
	try:
		cPickle.dump(data, open(filename, 'wb'))
		return True
	except Exception as e:
		log.error(e)
		return False
	
# read a text file
def read_file (filename):
	fid = open(filename, 'r')
	data = []
	for line in fid:			
		if len(line) > 0:
			line
			data.append (line.strip())		
	fid.close()
	return data

# read an dictionary from a json file
def read_JSON (filename, encoding = 'utf-8'):
	try:
		return json.load(open(filename, 'r'), encoding = encoding)
	except Exception as e:
		log.error(e)
		return None


# write a dictionary to a json file
def write_JSON (filename, data, encoding = 'utf-8'):
	try:
		json.dump(data, open(filename, 'w'), sort_keys = True, indent = 4, encoding = encoding)
		return True
	except Exception as e:
		log.error(e)
		return False

# read data from a csv file    
def read_csv (filename):
	try:
		reader = csv.reader (open(filename, "r"), escapechar='\\')
		data = []
		for r in reader:
			data.append(r)
		return data
	except Exception as e:
		log.error(e)
		return False

# read data from a csv file    
def read_csv_universal_newline (filename, delimiter = ","):
	try:
		reader = csv.reader (open(filename, "rU"), escapechar='\\', delimiter = delimiter)
		data = []
		for r in reader:
			data.append(r)
		return data
	except Exception as e:
		log.error(e)
		return False

# read a text file
def read_file (filename):
	try:
		fid = open(filename, 'r')
		data = []
		for line in fid:			
			if len(line) > 0:
				data.append (line.strip())		
		fid.close()
		return data
	except Exception as e:
		log.error(e)
		return False


# write data to a csv file
def write_csv (filename, data, quotechar='"', append = False, delimiter = ',', pivot = False):
	if quotechar is None: quoting=csv.QUOTE_NONE
	else: quoting=csv.QUOTE_ALL
	if append: writetype = 'a'
	else: writetype = 'wb'
	if pivot: data = zip(*data)
	try:
		doc = csv.writer (open(filename, writetype), delimiter=delimiter, quotechar=quotechar, quoting=quoting, escapechar = '\\')
		for d in data:
			doc.writerow (d)
		return True			
	except Exception as e:
		log.error(e)
		return False
	
# read a csv and output a list of dictionaries
def csvToDictionaryList(filename, delimiter = ','):
	try:
		reader = csv.DictReader (open(filename, "r"),delimiter = delimiter, escapechar='\\')
		data = []  
		for r in reader:
			data.append(r)
		return data
	except Exception as e:
		log.error(e)
		return False

# read a csv and output a dictionary with the keys as the specified column, and a list of dictionary rows for their result
def csvToDictionary(filename, key):
	try:
		reader = csv.DictReader (open(filename, "r"))
		data = {}
		for r in reader:
			keyvalue = r[key]
			if keyvalue in data:
				del r[key]
				data[keyvalue].append(r)
			else:
				del r[key]
				data[keyvalue] =[r]
		return data
	except Exception as e:
		log.error(e)
		return False

def dictionaryListToCSV(filename, data, fieldnames, encode = None, delimiter = ','):
#	try:
		doc = csv.DictWriter (open(filename, 'wb'), fieldnames=fieldnames, delimiter=delimiter, quotechar='"', quoting=csv.QUOTE_ALL, extrasaction = 'ignore')
		doc.writeheader()
		for d in data:
			if encode:
				for key, value in d.items():
					try:
						d[key] = value.encode(encode)
					except AttributeError:
						try:
							d[key] = ';'.join(value).encode(encode)
						except Exception:
							d[key] = '' 
			doc.writerow (d)
		return True			
# 	except Exception as e:
# 		log.error(e)
# 		return False

def prettyPrint(localobject, prefix = ''):
	if len(prefix) > 40:
		print prefix+json.dumps(localobject, sort_keys=True,
	                  	 indent=4, separators=(',', ': '))
	else:
		print prefix
		print json.dumps(localobject, sort_keys=True,
	                  	 indent=4, separators=(',', ': '))